@@ -2,6 +2,8 @@ from datetime import datetime, timezone
 from typing import Any, Tuple, Optional
 from uuid import uuid4
 
+import jwt
+
 from qottoauth import Namespace, Matching
 from qottoauth.api.base import QottoAuthApi, QottoAuthApiError
 from qottoauth.models import (
@@ -40,17 +42,41 @@ class QottoAuthTestApi(QottoAuthApi):
         self._roles: dict[str, Role] = {}
         self._role_authorizations: dict[str, set[str]] = {}
         self._members: dict[str, Member] = {}
-        self._member_cookies: dict[str, tuple[str, str]] = {}
         self._member_roles: dict[str, set[str]] = {}
         self._member_authorizations: dict[str, set[str]] = {}
         self._organizations: dict[str, Organization] = {}
         self._users: dict[str, User] = {}
-        self._user_cookies: dict[str, tuple[str, str]] = {}
         self._identities: dict[str, Identity] = {}
         self._identity_tickets: dict[str, str] = {}
         self._user_requests: dict[str, UserRequest] = {}
         self._last_change_date: datetime = datetime.now(timezone.utc)
 
+        self._token_cookie_name = 'token'
+        self._secret_cookie_name = 'secret'
+        self._token_algorithm = 'RS256'
+        self._secret_algorithm = 'RS256'
+        self._token_public_key = '''-----BEGIN RSA PUBLIC KEY-----
+MIGJAoGBAMvJcGqEd7hZL+LFBehqj73rqJOpT5A/yWBfcrpTTgq7hV0rthGHLiQM
+i5U7zjT+nbxYmCytig499egpibTPCnU/0SqXmlx4jelQb4QK30/TdktESlhgb5dg
+2qC5IhBrii/pb36wO+Tjz3t1F9EqWxnMRje3qdhVOJr/J9cdl/GDAgMBAAE=
+-----END RSA PUBLIC KEY-----'''
+        self._token_private_key = '''-----BEGIN RSA PRIVATE KEY-----
+MIICXAIBAAKBgQDLyXBqhHe4WS/ixQXoao+966iTqU+QP8lgX3K6U04Ku4VdK7YR
+hy4kDIuVO840/p28WJgsrYoOPfXoKYm0zwp1P9Eql5pceI3pUG+ECt9P03ZLREpY
+YG+XYNqguSIQa4ov6W9+sDvk4897dRfRKlsZzEY3t6nYVTia/yfXHZfxgwIDAQAB
+AoGALSTEqHQL6WSofP3UXzYr/ZCU2ZEqNPRMcfLzAV+u7CW6V3d1b9CYYhf9W+PR
+vF+jJbQikdMnwBvtu65n+QvcbtMKpXcmcRuaHqs4x7IoSxNKb1ILz+do1ekrFc16
+5oPd0EvGLyFcjH7O61He0iY3Uf/tRcKiCUPGU4BLbvLocZECQQD/ttxNFGCxxv2/
+ojiS4Ii1ODJ8QtbdKBp2fKUOf8XgvgHgt5KuvVW8kf6lBMqWcwNt85QskTDNLtCI
+r9dxJcuHAkEAzAO57aIEEDc0Qvp0gKK+34MRQtfBgAWNLF5amBtKdUd4B9QMdYbr
+fhOovbKtVca4ra93SRS01InkTV3hKXoBJQJAbaE3B7DB19XpOfxRZt7unUrvkgiR
+15T262960CGFc1nisjXhpBq2JDcvRg4s0J2UjdIM56KDmqQEcWV55x9+BwJAYVrd
+3OJVog9V5yhxc/k1sJ9xGz6uXhNIHQYhoThUvcaPJt0v3N23fwCOo4eiY65i7q8u
+8zboXAw5YBoOmqZX6QJBANyXY4RBSZC1xxJiWB6YGezRf4H4jR7+4ZfYDNfjBdcv
+IuFD3aYARUKtdS1jZW1Si5Bbhw7lLxxm9qgcmlZq00A=
+-----END RSA PRIVATE KEY-----'''
+        self._secret_private_key = self._token_private_key
+
     def query(
             self,
             name: str,
@@ -85,6 +111,14 @@ class QottoAuthTestApi(QottoAuthApi):
                 raise QottoAuthApiError from e
         raise QottoAuthApiError(f"Unknown mutation {name}")
 
+    def auth_metadata(self) -> dict[str, str]:
+        return dict(
+            token_cookie_name=self._token_cookie_name,
+            token_public_key=self._token_public_key,
+            token_algorithm=self._token_algorithm,
+            secret_cookie_name=self._secret_cookie_name,
+        )
+
     def lastChangeDate(self) -> str:
         return self._last_change_date.isoformat()
 
@@ -238,11 +272,10 @@ class QottoAuthTestApi(QottoAuthApi):
             nodes.append(user.to_dict())
         return dict(edges=[dict(node=node) for node in nodes])
 
-    def createUser(self, name: str, email: str, uuid: str = None, **kwargs) -> dict:
+    def createUser(self, name: str, uuid: str = None, **kwargs) -> dict:
         uuid = uuid or gen_uuid()
         user = User(uuid, name, False)
         self._users[user.id] = user
-        self._user_cookies[user.id] = str(uuid4()), str(uuid4())
         return dict(user=user.to_dict())
 
     def updateUser(
@@ -266,23 +299,18 @@ class QottoAuthTestApi(QottoAuthApi):
         identity = self._identities[identityId]
         user = User(gen_uuid(), identity.name, False)
         self._users[user.id] = user
-        self._user_cookies[user.id] = str(uuid4()), str(uuid4())
         self._identities[identityId] = identity.update(user=user)
         return dict(user=user.to_dict())
 
     def deleteUser(self, userId: str) -> dict:
         if userId in self._users:
             del self._users[userId]
-            del self._user_cookies[userId]
             for member in list(self._members.values()):
                 if member.user.id == userId:
                     self.deleteMember(member.id)
             for account in list(self._accounts.values()):
                 if account.user.id == userId:
                     self.deleteAccount(account.id)
-            for user_request in list(self._user_requests.values()):
-                if user_request.user.id == userId:
-                    self.deleteUserRequest(user_request.id)
             return dict(deleted=True)
         return dict(deleted=False)
 
@@ -323,7 +351,7 @@ class QottoAuthTestApi(QottoAuthApi):
         if identity_uuid not in self._identities:
             self._identities[identity_uuid] = Identity(
                 id=identity_uuid, name=identity_name, provider_id=providerId,
-                email=identity_email, user=None, blocked=False
+                email=identity_email, user=None, blocked=False, user_request=None,
             )
             self._identity_tickets[identity_uuid] = str(uuid4())
         return dict(identity=self._identities[identity_uuid].to_dict())
@@ -337,11 +365,11 @@ class QottoAuthTestApi(QottoAuthApi):
 
     def deleteIdentity(self, identityId: str) -> dict:
         if identityId in self._identities:
+            user_request = self._identities[identityId].user_request
+            if user_request and user_request.id in self._user_requests:
+                del self._user_requests[user_request.id]
             del self._identities[identityId]
             del self._identity_tickets[identityId]
-            for user_request in list(self._user_requests.values()):
-                if user_request.identity.id == identityId:
-                    self.deleteUserRequest(user_request.id)
             return dict(deleted=True)
         return dict(deleted=False)
 
@@ -473,7 +501,6 @@ class QottoAuthTestApi(QottoAuthApi):
         organization = self._organizations[organizationId]
         member = Member(gen_uuid(), user, organization)
         self._members[member.id] = member
-        self._member_cookies[member.id] = str(uuid4()), str(uuid4())
         self._member_authorizations[member.id] = set()
         self._member_roles[member.id] = set()
         return dict(member=member.to_dict())
@@ -498,7 +525,6 @@ class QottoAuthTestApi(QottoAuthApi):
             del self._members[memberId]
             del self._member_authorizations[memberId]
             del self._member_roles[memberId]
-            del self._member_cookies[memberId]
             return dict(deleted=True)
         return dict(deleted=False)
 
@@ -543,15 +569,13 @@ class QottoAuthTestApi(QottoAuthApi):
             return dict(deleted=True)
         return dict(deleted=False)
 
-    def createUserRequest(self, userId: str, identityId: str, comment: str = None) -> dict:
-        user = self._users[userId]
+    def createUserRequest(self, identityId: str, comment: str = None, **kwargs) -> dict:
         identity = self._identities[identityId]
-        if self.userRequests(userId=userId)['edges']:
-            raise QottoAuthApiError(f"User request for user {userId} already exists")
         if self.userRequests(identityId=identityId)['edges']:
             raise QottoAuthApiError(f"User request for identity {identityId} already exists")
-        request = UserRequest(id=gen_uuid(), user=user, identity=identity, comment=comment)
+        request = UserRequest(id=gen_uuid(), name=kwargs.pop('name', identity.name), comment=comment)
         self._user_requests[request.id] = request
+        self._identities[identityId] = identity.update(user_request=request)
         return dict(userRequest=request.to_dict())
 
     def userRequest(self, id: str) -> dict:
@@ -559,13 +583,18 @@ class QottoAuthTestApi(QottoAuthApi):
             return self._user_requests[id].to_dict()
         raise QottoAuthApiError(f"User request {id} does not exists")
 
-    def userRequests(self, userId: str = None, identityId: str = None) -> dict:
+    def userRequests(self, identityId: str = None) -> dict:
         nodes = []
         for request in self._user_requests.values():
-            if userId is not None and userId != request.user.id:
-                continue
-            if identityId is not None and identityId != request.identity.id:
-                continue
+            if identityId is not None:
+                for identity in self._identities.values():
+                    if identityId == identity.id:
+                        keep = bool(identity.user_request and identity.user_request.id == request.id)
+                        break
+                else:
+                    keep = False
+                if not keep:
+                    continue
             nodes.append(request.to_dict())
         return dict(edges=[dict(node=node) for node in nodes])
 
@@ -575,6 +604,44 @@ class QottoAuthTestApi(QottoAuthApi):
             return dict(deleted=True)
         return dict(deleted=False)
 
+    def acceptUserRequest(self, userRequestId: str) -> dict:
+        request = self._user_requests[userRequestId]
+        del self._user_requests[userRequestId]
+        for i in self._identities.values():
+            if i.user_request and i.user_request.id == request.id:
+                identity = i
+                break
+        else:
+            raise QottoAuthApiError(f"Identity for user request {userRequestId} does not exists")
+        user_data = self.createUser(name=request.name, identityId=identity.id)
+        user = self._users[user_data['user']['id']]
+        self.updateUser(user.id, addIdentityId=identity.id)
+        identity = self._identities[identity.id]
+        return dict(user=user.to_dict(), identity=identity.to_dict())
+
+    def rejectUserRequest(
+            self, userRequestId: str,
+            blockIdentity: bool = False,
+            deleteIdentity: bool = False,
+    ) -> dict:
+        request = self._user_requests[userRequestId]
+        for i in self._identities.values():
+            if i.user_request and i.user_request.id == request.id:
+                identity = i
+                break
+        else:
+            raise QottoAuthApiError(f"Identity for user request {userRequestId} does not exists")
+        del self._user_requests[userRequestId]
+        if blockIdentity:
+            self.updateIdentity(identity.id, setBlocked=True)
+        if deleteIdentity:
+            self.deleteIdentity(identity.id)
+
+        opt_identity = self._identities.get(identity.id)
+        return dict(
+            identity=opt_identity.to_dict() if opt_identity else None,
+        )
+
     def identityTicket(self, identityId: str) -> str:
         if identityId in self._identity_tickets:
             return self._identity_tickets[identityId]
@@ -587,34 +654,83 @@ class QottoAuthTestApi(QottoAuthApi):
         raise QottoAuthApiError(f"Identity ticket {identityTicket} does not exists")
 
     def actor(self, tokenCookie: str = None, secretCookie: str = None) -> dict:
-        for member_id, cookies in self._member_cookies.items():
-            if cookies == (tokenCookie, secretCookie):
-                member = self._members[member_id]
-                actor = Actor(user=member.user, member=member)
-                return actor.to_dict()
-        for user_id, cookies in self._user_cookies.items():
-            if cookies == (tokenCookie, secretCookie):
-                user = self._users[user_id]
-                actor = Actor(user=user)
-                return actor.to_dict()
-        return Actor().to_dict()
+        if not tokenCookie or not secretCookie:
+            return Actor().to_dict()
+        payload = jwt.decode(tokenCookie, self._token_public_key, algorithms=[self._token_algorithm])
+        computed_secret = jwt.encode(
+            dict(token=tokenCookie), self._secret_private_key, algorithm=self._secret_algorithm
+            )
+        if computed_secret != secretCookie:
+            return Actor().to_dict()
+
+        user = self._users.get(payload.get('sub', ''))
+        if user is None:
+            return Actor().to_dict()
+
+        member = self._members.get(payload.get('metadata', {}).get('member_id', ''))
+        if member is None:
+            return Actor(user=user).to_dict()
+
+        return Actor(user=user, member=member).to_dict()
 
     def cookies(self, userId: str = None, organizationId: str = None) -> dict:
+        user: Optional[User] = None
+        member: Optional[Member] = None
         if userId and organizationId:
             for member in self._members.values():
                 if member.organization.id == organizationId and member.user.id == userId:
-                    token, secret = self._member_cookies[member.id]
+                    user = member.user
+                    member = member
                     break
             else:
                 raise QottoAuthApiError(f"Member {userId} does not exists in organization {organizationId}")
         elif userId:
-            if userId in self._user_cookies:
-                token, secret = self._user_cookies[userId]
+            if userId in self._users:
+                user = self._users[userId]
             else:
                 raise QottoAuthApiError(f"User {userId} does not exists")
-        else:
-            token, secret = '', ''
-        cookie_pair = CookiePair(
+
+        if not user:
+            return CookiePair(
+                token_cookie=Cookie(
+                    name='token', value='', domain='localhost',
+                    max_age=0, secure=False, http_only=False,
+                ),
+                secret_cookie=Cookie(
+                    name='secret', value='', domain='localhost',
+                    max_age=0, secure=False, http_only=True,
+                ),
+            ).to_dict()
+
+        payload: dict[str, Any] = dict(
+            iss='test',
+            sub=user.id,
+            uuid=user.id,
+            is_superuser=user.is_superuser,
+            metadata=dict(
+                displayed_name=user.name,
+                given_name='given',
+                middle_name='middle',
+                family_name='family',
+                email='email',
+            ),
+            applications=[
+                account.application.name
+                for account in self._accounts.values()
+                if account.user.id == user.id and account.enabled
+            ],
+        )
+        if member:
+            payload['org'] = member.organization.id
+            payload['metadata']['member_id'] = member.id
+            payload['metadata']['organization_id'] = member.organization.id
+            payload['metadata']['organization_name'] = member.organization.name
+            payload['metadata']['organization_namespace'] = str(member.organization.namespace)
+
+        token = jwt.encode(payload, self._token_private_key, algorithm=self._token_algorithm)
+        secret = jwt.encode(dict(token=token), self._secret_private_key, algorithm=self._token_algorithm)
+
+        return CookiePair(
             token_cookie=Cookie(
                 name='token', value=token, domain='localhost',
                 max_age=0, secure=False, http_only=False,
@@ -623,8 +739,7 @@ class QottoAuthTestApi(QottoAuthApi):
                 name='secret', value=secret, domain='localhost',
                 max_age=0, secure=False, http_only=True,
             ),
-        )
-        return cookie_pair.to_dict()
+        ).to_dict()
 
     def isAuthorized(
             self,
